@@ -2,7 +2,7 @@
  *
  *  OBEX Server
  *
- *  Copyright (C) 2007-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2007-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -40,9 +40,10 @@
 #include "bluetooth.h"
 #include "obexd.h"
 #include "obex.h"
+#include "dbus.h"
 #include "logging.h"
 
-const static gchar *opp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
+static const gchar *opp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
 <record>									\
   <attribute id=\"0x0001\">							\
     <sequence>									\
@@ -92,7 +93,7 @@ const static gchar *opp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
   </attribute>									\
 </record>";
 
-const static gchar *ftp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
+static const gchar *ftp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
 <record>									\
   <attribute id=\"0x0001\">							\
     <sequence>									\
@@ -129,7 +130,7 @@ const static gchar *ftp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
   </attribute>									\
 </record>";
 
-const static gchar *pbap_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
+static const gchar *pbap_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
 <record>									\
   <attribute id=\"0x0001\">							\
     <sequence>									\
@@ -170,6 +171,49 @@ const static gchar *pbap_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
   </attribute>									\
 </record>";
 
+static const gchar *pcsuite_record =
+"<?xml version=\"1.0\" encoding=\"UTF-8\" ?>					\
+<record>									\
+  <attribute id=\"0x0001\">							\
+    <sequence>									\
+      <uuid value=\"00005005-0000-1000-8000-0002ee000001\"/>			\
+    </sequence>									\
+  </attribute>									\
+										\
+  <attribute id=\"0x0004\">							\
+    <sequence>									\
+      <sequence>								\
+        <uuid value=\"0x0100\"/>						\
+      </sequence>								\
+      <sequence>								\
+        <uuid value=\"0x0003\"/>						\
+        <uint8 value=\"%u\" name=\"channel\"/>					\
+      </sequence>								\
+      <sequence>								\
+        <uuid value=\"0x0008\"/>						\
+      </sequence>								\
+    </sequence>									\
+  </attribute>									\
+										\
+  <attribute id=\"0x0005\">							\
+    <sequence>									\
+      <uuid value=\"0x1002\"/>							\
+    </sequence>									\
+  </attribute>									\
+										\
+  <attribute id=\"0x0009\">							\
+    <sequence>									\
+      <sequence>								\
+        <uuid value=\"00005005-0000-1000-8000-0002ee000001\"/> 			\
+        <uint16 value=\"0x0100\" name=\"version\"/>				\
+      </sequence>								\
+    </sequence>									\
+  </attribute>									\
+										\
+  <attribute id=\"0x0100\">							\
+    <text value=\"%s\" name=\"name\"/>						\
+  </attribute>									\
+</record>";
 
 #define TRANSFER_INTERFACE OPENOBEX_SERVICE ".Transfer"
 #define SESSION_INTERFACE OPENOBEX_SERVICE ".Session"
@@ -188,6 +232,7 @@ static struct agent *agent = NULL;
 
 struct pending_request {
 	struct server *server;
+	gchar address[18];
 	gint nsk;
 };
 
@@ -473,6 +518,9 @@ static gchar *create_xml_record(const char *name,
 	case OBEX_PBAP:
 		xml = g_markup_printf_escaped(pbap_record, channel, name);
 		break;
+	case OBEX_PCSUITE:
+		xml = g_markup_printf_escaped(pcsuite_record, channel, name);
+		break;
 	default:
 		xml = NULL;
 		break;
@@ -534,9 +582,8 @@ failed:
 	return ret;
 }
 
-void register_record(gpointer data, gpointer user_data)
+void register_record(struct server *server, gpointer user_data)
 {
-	struct server *server = data;
 	gchar *xml;
 	gint ret;
 
@@ -554,7 +601,7 @@ void register_record(gpointer data, gpointer user_data)
 	g_free(xml);
 }
 
-static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
+static void find_adapter_any_reply(DBusPendingCall *call, gpointer user_data)
 {
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	struct server *server;
@@ -591,12 +638,14 @@ done:
 
 	dbus_message_unref(reply);
 }
-
-static gboolean find_adapter_any(gpointer user_data)
+static void find_adapter(const char *pattern,
+				DBusPendingCallNotifyFunction function,
+				gpointer user_data)
 {
 	DBusMessage *msg;
 	DBusPendingCall *call;
-	const char *pattern = "any";
+
+	debug("FindAdapter(%s)", pattern);
 
 	msg = dbus_message_new_method_call("org.bluez", "/",
 					"org.bluez.Manager", "FindAdapter");
@@ -605,10 +654,15 @@ static gboolean find_adapter_any(gpointer user_data)
 			DBUS_TYPE_INVALID);
 
 	dbus_connection_send_with_reply(system_conn, msg, &call, -1);
-	dbus_pending_call_set_notify(call, find_adapter_reply, NULL, NULL);
+	dbus_pending_call_set_notify(call, function, user_data, NULL);
 	dbus_pending_call_unref(call);
 
 	dbus_message_unref(msg);
+}
+
+static gboolean find_adapter_any(gpointer user_data)
+{
+	find_adapter("any", find_adapter_any_reply, user_data);
 
 	return FALSE;
 }
@@ -800,6 +854,7 @@ static void agent_reply(DBusPendingCall *call, gpointer user_data)
 				DBUS_TYPE_INVALID)) {
 		/* Splits folder and name */
 		const gchar *slash = strrchr(name, '/');
+		debug("Agent replied with %s", name);
 		if (!slash) {
 			agent->new_name = g_strdup(name);
 			agent->new_folder = NULL;
@@ -856,7 +911,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	addrlen = sizeof(addr);
 
 	if (getpeername(fd, (struct sockaddr *) &addr, &addrlen) < 0)
-		return -1;
+		return -errno;
 
 	ba2str(&addr.rc_bdaddr, address);
 
@@ -923,21 +978,20 @@ static void service_reply(DBusPendingCall *call, gpointer user_data)
 
 	dbus_error_init(&derr);
 	if (dbus_set_error_from_message(&derr, reply)) {
-		error("Replied with an error: %s, %s",
+		error("RequestAuthorization error: %s, %s",
 				derr.name, derr.message);
 		dbus_error_free(&derr);
-		dbus_message_unref(reply);
-
 		close(pending->nsk);
-		g_free(pending);
-		return;
+		goto done;
 	}
 
+	debug("RequestAuthorization succeeded");
+
 	if (obex_session_start(pending->nsk, pending->server) < 0)
 		close(pending->nsk);
 
+done:
 	g_free(pending);
-
 	dbus_message_unref(reply);
 }
 
@@ -961,16 +1015,73 @@ static gboolean service_error(GIOChannel *io, GIOCondition cond,
 	return FALSE;
 }
 
+static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
+{
+	struct pending_request *pending = user_data;
+	DBusMessage *reply = dbus_pending_call_steal_reply(call);
+	DBusMessage *msg;
+	DBusPendingCall *pcall;
+	GIOChannel *io;
+	const char *paddr = pending->address;
+	const char *path;
+	DBusError derr;
+
+	dbus_error_init(&derr);
+	if (dbus_set_error_from_message(&derr, reply)) {
+		error("Replied with an error: %s, %s",
+				derr.name, derr.message);
+		dbus_error_free(&derr);
+		goto done;
+	}
+
+	dbus_message_get_args(reply, NULL,
+			DBUS_TYPE_OBJECT_PATH, &path,
+			DBUS_TYPE_INVALID);
+
+	debug("FindAdapter -> %s", path);
+
+	msg = dbus_message_new_method_call("org.bluez", path,
+			"org.bluez.Service", "RequestAuthorization");
+
+	dbus_message_append_args(msg, DBUS_TYPE_STRING, &paddr,
+			DBUS_TYPE_UINT32, &pending->server->handle,
+			DBUS_TYPE_INVALID);
+
+	if (!dbus_connection_send_with_reply(system_conn,
+					msg, &pcall, TIMEOUT)) {
+		dbus_message_unref(msg);
+		goto done;
+	}
+
+	dbus_message_unref(msg);
+
+	debug("RequestAuthorization(%s, %x)", paddr, pending->server->handle);
+
+	if (!dbus_pending_call_set_notify(pcall, service_reply, pending, NULL)) {
+		close(pending->nsk);
+		g_free(pending);
+		goto done;
+	}
+
+	dbus_pending_call_unref(pcall);
+
+	/* Catches errors before authorization response comes */
+	io = g_io_channel_unix_new(pending->nsk);
+	g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
+			G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+			service_error, NULL, NULL);
+	g_io_channel_unref(io);
+
+done:
+	dbus_message_unref(reply);
+}
+
 gint request_service_authorization(struct server *server, gint nsk)
 {
-	DBusMessage *msg;
-	DBusPendingCall *call;
-	GIOChannel *io;
 	struct sockaddr_rc src, dst;
 	socklen_t addrlen;
 	gchar address[18];
 	const gchar *paddr = address;
-	guint watch;
 	struct pending_request *pending;
 
 	if (system_conn == NULL || any->path == NULL)
@@ -988,41 +1099,13 @@ gint request_service_authorization(struct server *server, gint nsk)
 	if (getpeername(nsk, (struct sockaddr *) &dst, &addrlen) < 0)
 		return -1;
 
-	ba2str(&dst.rc_bdaddr, address);
-
-	msg = dbus_message_new_method_call("org.bluez", any->path,
-			"org.bluez.Service", "RequestAuthorization");
-
-	dbus_message_append_args(msg, DBUS_TYPE_STRING, &paddr,
-			DBUS_TYPE_UINT32, &server->handle,
-			DBUS_TYPE_INVALID);
-
-	if (!dbus_connection_send_with_reply(system_conn,
-					msg, &call, TIMEOUT)) {
-		dbus_message_unref(msg);
-		return -EPERM;
-	}
-
-	dbus_message_unref(msg);
-
 	pending = g_new0(struct pending_request, 1);
 	pending->server = server;
 	pending->nsk = nsk;
 
-	if (!dbus_pending_call_set_notify(call, service_reply, pending, NULL)) {
-		close(nsk);
-		g_free(pending);
-		return -EPERM;
-	}
+	ba2str(&dst.rc_bdaddr, pending->address);
 
-	dbus_pending_call_unref(call);
-
-	/* Catches errors before authorization response comes */
-	io = g_io_channel_unix_new(nsk);
-	watch = g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
-			G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-			service_error, NULL, NULL);
-	g_io_channel_unref(io);
+	find_adapter(paddr, find_adapter_reply, pending);
 
 	return 0;
 }
